@@ -11,7 +11,7 @@
 ******************************************************************************/
 
 
-#ifdef __MSDOS__
+#if defined (__MSDOS__) && !defined(__DJGPP__) && !defined(__GNUC__)
 #include <io.h>
 #include <alloc.h>
 #include <stdlib.h>
@@ -21,75 +21,37 @@
 #include <sys/stat.h>
 #endif /* __MSDOS__ */
 
+#ifndef __MSDOS__
+#include <stdlib.h>
+#endif
 #include <fcntl.h>
 #include <stdio.h>
 #include <string.h>
 #include "gif_lib.h"
-#include "gif_hash.h"
-
-#define PROGRAM_NAME	"GIFLIB"
+#include "gif_lib_private.h"
 
 #define COMMENT_EXT_FUNC_CODE	0xfe /* Extension function code for comment. */
 #define GIF_STAMP	"GIFVER"	 /* First chars in file - GIF stamp. */
 #define GIF_STAMP_LEN	sizeof(GIF_STAMP) - 1
 #define GIF_VERSION_POS	3		/* Version first character in stamp. */
 
-#define LZ_MAX_CODE	4095		/* Biggest code possible in 12 bits. */
-#define LZ_BITS		12
-
 #define FILE_STATE_READ		0x01/* 1 write, 0 read - EGIF_LIB compatible.*/
 
-#define FLUSH_OUTPUT		4096    /* Impossible code, to signal flush. */
-#define FIRST_CODE		4097    /* Impossible code, to signal first. */
-#define NO_SUCH_CODE		4098    /* Impossible code, to signal empty. */
-
 #define IS_READABLE(Private)	(!(Private->FileState & FILE_STATE_READ))
 
-typedef struct GifFilePrivateType {
-    int FileState,
-	FileHandle,			     /* Where all this data goes to! */
-	BitsPerPixel,	    /* Bits per pixel (Codes uses at list this + 1). */
-	ClearCode,				       /* The CLEAR LZ code. */
-	EOFCode,				         /* The EOF LZ code. */
-	RunningCode,		    /* The next code algorithm can generate. */
-	RunningBits,/* The number of bits required to represent RunningCode. */
-	MaxCode1,  /* 1 bigger than max. possible code, in RunningBits bits. */
-	LastCode,		        /* The code before the current code. */
-	CrntCode,				  /* Current algorithm code. */
-	StackPtr,		         /* For character stack (see below). */
-	CrntShiftState;		        /* Number of bits in CrntShiftDWord. */
-    unsigned long CrntShiftDWord;     /* For bytes decomposition into codes. */
-    unsigned long PixelCount;		       /* Number of pixels in image. */
-    FILE *File;						  /* File as stream. */
-    GifByteType Buf[256];	       /* Compressed input is buffered here. */
-    GifByteType Stack[LZ_MAX_CODE];	 /* Decoded pixels are stacked here. */
-    GifByteType Suffix[LZ_MAX_CODE+1];	       /* So we can trace the codes. */
-    unsigned int Prefix[LZ_MAX_CODE+1];
-} GifFilePrivateType;
+/* avoid extra function call in case we use fread (TVT) */
+#define READ(_gif,_buf,_len)                                     \
+  (((GifFilePrivateType*)_gif->Private)->Read ?                   \
+    ((GifFilePrivateType*)_gif->Private)->Read(_gif,_buf,_len) : \
+    fread(_buf,1,_len,((GifFilePrivateType*)_gif->Private)->File))
 
-#ifdef SYSV
-static char *VersionStr =
-        "Gif library module,\t\tGershon Elber\n\
-	(C) Copyright 1989 Gershon Elber, Non commercial use only.\n";
-#else
-static char *VersionStr =
-	PROGRAM_NAME
-	"	IBMPC "
-	GIF_LIB_VERSION
-	"	Gershon Elber,	"
-	__DATE__ ",   " __TIME__ "\n"
-	"(C) Copyright 1989 Gershon Elber, Non commercial use only.\n";
-#endif /* SYSV */
-
-extern int _GifError;
-
-static int DGifGetWord(FILE *File, int *Word);
+static int DGifGetWord(GifFileType *GifFile, int *Word);
 static int DGifSetupDecompress(GifFileType *GifFile);
 static int DGifDecompressLine(GifFileType *GifFile, GifPixelType *Line,
 								int LineLen);
 static int DGifGetPrefixChar(unsigned int *Prefix, int Code, int ClearCode);
-static int DGifDecompressInput(GifFilePrivateType *Private, int *Code);
-static int DGifBufferedInput(FILE *File, GifByteType *Buf,
+static int DGifDecompressInput(GifFileType *GifFile, int *Code);
+static int DGifBufferedInput(GifFileType *GifFile, GifByteType *Buf,
 						     GifByteType *NextByte);
 
 /******************************************************************************
@@ -150,9 +112,11 @@ GifFileType *DGifOpenFileHandle(int FileHandle)
     Private->FileHandle = FileHandle;
     Private->File = f;
     Private->FileState = 0;   /* Make sure bit 0 = 0 (File open for read). */
+	Private->Read     = 0;  /* don't use alternate input method (TVT) */
+	GifFile->UserData = 0;  /* TVT */
 
     /* Lets see if this is a GIF file: */
-    if (fread(Buf, 1, GIF_STAMP_LEN, Private->File) != GIF_STAMP_LEN) {
+    if (READ(GifFile,Buf, GIF_STAMP_LEN) != GIF_STAMP_LEN) {
 	_GifError = D_GIF_ERR_READ_FAILED;
 	free((char *) Private);
 	free((char *) GifFile);
@@ -180,6 +144,67 @@ GifFileType *DGifOpenFileHandle(int FileHandle)
     return GifFile;
 }
 
+/******************************************************************************
+* GifFileType constructor with user supplied input function (TVT)             *
+*                                                                             *
+******************************************************************************/
+GifFileType *DGifOpen( void* userData, InputFunc readFunc )
+{
+    char Buf[GIF_STAMP_LEN+1];
+    GifFileType *GifFile;
+    GifFilePrivateType *Private;
+
+
+    if ((GifFile = (GifFileType *) malloc(sizeof(GifFileType))) == NULL) {
+	  _GifError = D_GIF_ERR_NOT_ENOUGH_MEM;
+	  return NULL;
+    }
+
+    memset(GifFile, '\0', sizeof(GifFileType));
+
+    if (!(Private = (GifFilePrivateType*) malloc(sizeof(GifFilePrivateType)))){
+	  _GifError = D_GIF_ERR_NOT_ENOUGH_MEM;
+	  free((char *) GifFile);
+	  return NULL;
+    }
+
+    GifFile->Private = (VoidPtr) Private;
+    Private->FileHandle = 0;
+    Private->File = 0;
+    Private->FileState = 0;   /* Make sure bit 0 = 0 (File open for read). */
+
+	Private->Read     = readFunc;  /* TVT */
+	GifFile->UserData = userData;    /* TVT */
+
+    /* Lets see if this is a GIF file: */
+    if ( READ( GifFile, Buf, GIF_STAMP_LEN) != GIF_STAMP_LEN) {
+	  _GifError = D_GIF_ERR_READ_FAILED;
+	  free((char *) Private);
+	  free((char *) GifFile);
+	  return NULL;
+    }
+
+    /* The GIF Version number is ignored at this time. Maybe we should do    */
+    /* something more useful with it.					     */
+    Buf[GIF_STAMP_LEN] = 0;
+    if (strncmp(GIF_STAMP, Buf, GIF_VERSION_POS) != 0) {
+	  _GifError = D_GIF_ERR_NOT_GIF_FILE;
+	  free((char *) Private);
+	  free((char *) GifFile);
+	  return NULL;
+    }
+
+    if (DGifGetScreenDesc(GifFile) == GIF_ERROR) {
+	  free((char *) Private);
+	  free((char *) GifFile);
+	  return NULL;
+    }
+
+    _GifError = 0;
+
+    return GifFile;
+}
+
 /******************************************************************************
 *   This routine should be called before any other DGif calls. Note that      *
 * this routine is called automatically from DGif file open routines.	      *
@@ -197,11 +222,11 @@ int DGifGetScreenDesc(GifFileType *GifFile)
     }
 
     /* Put the screen descriptor into the file: */
-    if (DGifGetWord(Private->File, &GifFile->SWidth) == GIF_ERROR ||
-	DGifGetWord(Private->File, &GifFile->SHeight) == GIF_ERROR)
+    if (DGifGetWord(GifFile, &GifFile->SWidth) == GIF_ERROR ||
+	DGifGetWord(GifFile, &GifFile->SHeight) == GIF_ERROR)
 	return GIF_ERROR;
 
-    if (fread(Buf, 1, 3, Private->File) != 3) {
+    if (READ( GifFile, Buf, 3) != 3) {
 	_GifError = D_GIF_ERR_READ_FAILED;
 	return GIF_ERROR;
     }
@@ -214,7 +239,7 @@ int DGifGetScreenDesc(GifFileType *GifFile)
 
 	/* Get the global color map: */
 	for (i = 0; i < GifFile->SColorMap->ColorCount; i++) {
-	    if (fread(Buf, 1, 3, Private->File) != 3) {
+	    if (READ(GifFile, Buf, 3) != 3) {
 		_GifError = D_GIF_ERR_READ_FAILED;
 		return GIF_ERROR;
 	    }
@@ -241,7 +266,7 @@ int DGifGetRecordType(GifFileType *GifFile, GifRecordType *Type)
 	return GIF_ERROR;
     }
 
-    if (fread(&Buf, 1, 1, Private->File) != 1) {
+    if (READ( GifFile, &Buf, 1) != 1) {
 	_GifError = D_GIF_ERR_READ_FAILED;
 	return GIF_ERROR;
     }
@@ -274,6 +299,7 @@ int DGifGetImageDesc(GifFileType *GifFile)
     int i, BitsPerPixel;
     GifByteType Buf[3];
     GifFilePrivateType *Private = (GifFilePrivateType *) GifFile->Private;
+	SavedImage	*sp;
 
     if (!IS_READABLE(Private)) {
 	/* This file was NOT open for reading: */
@@ -281,12 +307,12 @@ int DGifGetImageDesc(GifFileType *GifFile)
 	return GIF_ERROR;
     }
 
-    if (DGifGetWord(Private->File, &GifFile->Image.Left) == GIF_ERROR ||
-	DGifGetWord(Private->File, &GifFile->Image.Top) == GIF_ERROR ||
-	DGifGetWord(Private->File, &GifFile->Image.Width) == GIF_ERROR ||
-	DGifGetWord(Private->File, &GifFile->Image.Height) == GIF_ERROR)
+    if (DGifGetWord(GifFile, &GifFile->Image.Left) == GIF_ERROR ||
+	DGifGetWord(GifFile, &GifFile->Image.Top) == GIF_ERROR ||
+	DGifGetWord(GifFile, &GifFile->Image.Width) == GIF_ERROR ||
+	DGifGetWord(GifFile, &GifFile->Image.Height) == GIF_ERROR)
 	return GIF_ERROR;
-    if (fread(Buf, 1, 1, Private->File) != 1) {
+    if (READ(GifFile,Buf, 1) != 1) {
 	_GifError = D_GIF_ERR_READ_FAILED;
 	return GIF_ERROR;
     }
@@ -301,7 +327,7 @@ int DGifGetImageDesc(GifFileType *GifFile)
     
 	/* Get the image local color map: */
 	for (i = 0; i < GifFile->Image.ColorMap->ColorCount; i++) {
-	    if (fread(Buf, 1, 3, Private->File) != 3) {
+	    if (READ(GifFile,Buf, 3) != 3) {
 		_GifError = D_GIF_ERR_READ_FAILED;
 		return GIF_ERROR;
 	    }
@@ -312,20 +338,34 @@ int DGifGetImageDesc(GifFileType *GifFile)
     }
 
     if (GifFile->SavedImages) {
-	SavedImage	*sp;
-
-	if ((GifFile->SavedImages = (SavedImage *)realloc(GifFile->SavedImages,
-		    sizeof(SavedImage) * (GifFile->ImageCount + 1))) == NULL) {
-	    _GifError = D_GIF_ERR_NOT_ENOUGH_MEM;
-	    return GIF_ERROR;
-	}
+	    if ((GifFile->SavedImages = (SavedImage *)realloc(GifFile->SavedImages,
+		     sizeof(SavedImage) * (GifFile->ImageCount + 1))) == NULL) {
+	        _GifError = D_GIF_ERR_NOT_ENOUGH_MEM;
+	        return GIF_ERROR;
+	    }
+    } else {
+        if ((GifFile->SavedImages =
+             (SavedImage *)malloc(sizeof(SavedImage))) == NULL) {
+            _GifError = D_GIF_ERR_NOT_ENOUGH_MEM;
+            return GIF_ERROR;
+        }
+    }
 
 	sp = &GifFile->SavedImages[GifFile->ImageCount];
 	memcpy(&sp->ImageDesc, &GifFile->Image, sizeof(GifImageDesc));
+    if (GifFile->Image.ColorMap != NULL) {
+	    sp->ImageDesc.ColorMap =
+               (ColorMapObject *)malloc(sizeof (ColorMapObject));
+	    memcpy(&sp->ImageDesc.ColorMap, &GifFile->Image.ColorMap,
+               sizeof(ColorMapObject));
+	    sp->ImageDesc.ColorMap->Colors =
+               (GifColorType *)malloc(sizeof (GifColorType));
+	    memcpy(&sp->ImageDesc.ColorMap->Colors,
+               &GifFile->Image.ColorMap->Colors, sizeof(GifColorType));
+    }
 	sp->RasterBits = (char *)NULL;
 	sp->ExtensionBlockCount = 0;
 	sp->ExtensionBlocks = (ExtensionBlock *)NULL;
-    }
 
     GifFile->ImageCount++;
 
@@ -435,7 +475,7 @@ int DGifGetExtension(GifFileType *GifFile, int *ExtCode,
 	return GIF_ERROR;
     }
 
-    if (fread(&Buf, 1, 1, Private->File) != 1) {
+    if (READ(GifFile,&Buf, 1) != 1) {
 	_GifError = D_GIF_ERR_READ_FAILED;
 	return GIF_ERROR;
     }
@@ -454,14 +494,14 @@ int DGifGetExtensionNext(GifFileType *GifFile, GifByteType **Extension)
     GifByteType Buf;
     GifFilePrivateType *Private = (GifFilePrivateType *) GifFile->Private;
 
-    if (fread(&Buf, 1, 1, Private->File) != 1) {
+    if (READ(GifFile,&Buf, 1) != 1) {
 	_GifError = D_GIF_ERR_READ_FAILED;
 	return GIF_ERROR;
     }
     if (Buf > 0) {
 	*Extension = Private->Buf;           /* Use private unused buffer. */
 	(*Extension)[0] = Buf;  /* Pascal strings notation (pos. 0 is len.). */
-	if (fread(&((*Extension)[1]), 1, Buf, Private->File) != Buf) {
+	if (READ(GifFile,&((*Extension)[1]), Buf) != Buf) {
 	    _GifError = D_GIF_ERR_READ_FAILED;
 	    return GIF_ERROR;
 	}
@@ -502,9 +542,9 @@ int DGifCloseFile(GifFileType *GifFile)
 	FreeSavedImages(GifFile);
     free(GifFile);
 
-    if (fclose(File) != 0) {
-	_GifError = D_GIF_ERR_CLOSE_FAILED;
-	return GIF_ERROR;
+    if ( File && (fclose(File) != 0)) {
+	  _GifError = D_GIF_ERR_CLOSE_FAILED;
+	  return GIF_ERROR;
     }
     return GIF_OK;
 }
@@ -512,11 +552,11 @@ int DGifCloseFile(GifFileType *GifFile)
 /******************************************************************************
 *   Get 2 bytes (word) from the given file:				      *
 ******************************************************************************/
-static int DGifGetWord(FILE *File, int *Word)
+static int DGifGetWord(GifFileType *GifFile, int *Word)
 {
     unsigned char c[2];
 
-    if (fread(c, 1, 2, File) != 2) {
+    if (READ(GifFile,c, 2) != 2) {
 	_GifError = D_GIF_ERR_READ_FAILED;
 	return GIF_ERROR;
     }
@@ -557,7 +597,7 @@ int DGifGetCodeNext(GifFileType *GifFile, GifByteType **CodeBlock)
     GifByteType Buf;
     GifFilePrivateType *Private = (GifFilePrivateType *) GifFile->Private;
 
-    if (fread(&Buf, 1, 1, Private->File) != 1) {
+    if (READ(GifFile,&Buf, 1) != 1) {
 	_GifError = D_GIF_ERR_READ_FAILED;
 	return GIF_ERROR;
     }
@@ -565,7 +605,7 @@ int DGifGetCodeNext(GifFileType *GifFile, GifByteType **CodeBlock)
     if (Buf > 0) {
 	*CodeBlock = Private->Buf;	       /* Use private unused buffer. */
 	(*CodeBlock)[0] = Buf;  /* Pascal strings notation (pos. 0 is len.). */
-	if (fread(&((*CodeBlock)[1]), 1, Buf, Private->File) != Buf) {
+	if (READ(GifFile,&((*CodeBlock)[1]), Buf) != Buf) {
 	    _GifError = D_GIF_ERR_READ_FAILED;
 	    return GIF_ERROR;
 	}
@@ -589,7 +629,7 @@ static int DGifSetupDecompress(GifFileType *GifFile)
     unsigned int *Prefix;
     GifFilePrivateType *Private = (GifFilePrivateType *) GifFile->Private;
 
-    fread(&CodeSize, 1, 1, Private->File);    /* Read Code size from file. */
+    READ(GifFile,&CodeSize, 1);    /* Read Code size from file. */
     BitsPerPixel = CodeSize;
 
     Private->Buf[0] = 0;			      /* Input Buffer empty. */
@@ -638,7 +678,7 @@ static int DGifDecompressLine(GifFileType *GifFile, GifPixelType *Line,
     }
 
     while (i < LineLen) {			    /* Decode LineLen items. */
-	if (DGifDecompressInput(Private, &CrntCode) == GIF_ERROR)
+	if (DGifDecompressInput(GifFile, &CrntCode) == GIF_ERROR)
     	    return GIF_ERROR;
 
 	if (CrntCode == EOFCode) {
@@ -769,7 +809,7 @@ int DGifGetLZCodes(GifFileType *GifFile, int *Code)
 	return GIF_ERROR;
     }
 
-    if (DGifDecompressInput(Private, Code) == GIF_ERROR)
+    if (DGifDecompressInput(GifFile, Code) == GIF_ERROR)
 	return GIF_ERROR;
 
     if (*Code == Private->EOFCode) {
@@ -796,8 +836,10 @@ int DGifGetLZCodes(GifFileType *GifFile, int *Code)
 * 8 bits (bytes) packets, into the real codes.				      *
 *   Returns GIF_OK if read succesfully.					      *
 ******************************************************************************/
-static int DGifDecompressInput(GifFilePrivateType *Private, int *Code)
+static int DGifDecompressInput(GifFileType *GifFile, int *Code)
 {
+    GifFilePrivateType *Private = (GifFilePrivateType *)GifFile->Private;
+
     GifByteType NextByte;
     static unsigned int CodeMasks[] = {
 	0x0000, 0x0001, 0x0003, 0x0007,
@@ -808,7 +850,7 @@ static int DGifDecompressInput(GifFilePrivateType *Private, int *Code)
 
     while (Private->CrntShiftState < Private->RunningBits) {
 	/* Needs to get more bytes from input stream for next code: */
-	if (DGifBufferedInput(Private->File, Private->Buf, &NextByte)
+	if (DGifBufferedInput(GifFile, Private->Buf, &NextByte)
 	    == GIF_ERROR) {
 	    return GIF_ERROR;
 	}
@@ -837,17 +879,17 @@ static int DGifDecompressInput(GifFilePrivateType *Private, int *Code)
 *   The routine returns the next byte from its internal buffer (or read next  *
 * block in if buffer empty) and returns GIF_OK if succesful.		      *
 ******************************************************************************/
-static int DGifBufferedInput(FILE *File, GifByteType *Buf,
+static int DGifBufferedInput(GifFileType *GifFile, GifByteType *Buf,
 						      GifByteType *NextByte)
 {
     if (Buf[0] == 0) {
 	/* Needs to read the next buffer - this one is empty: */
-	if (fread(Buf, 1, 1, File) != 1)
+	if (READ(GifFile, Buf, 1) != 1)
 	{
 	    _GifError = D_GIF_ERR_READ_FAILED;
 	    return GIF_ERROR;
 	}
-	if (fread(&Buf[1], 1, Buf[0], File) != Buf[0])
+	if (READ(GifFile,&Buf[1], Buf[0]) != Buf[0])
 	{
 	    _GifError = D_GIF_ERR_READ_FAILED;
 	    return GIF_ERROR;
@@ -868,18 +910,19 @@ static int DGifBufferedInput(FILE *File, GifByteType *Buf,
 * This routine reads an entire GIF into core, hanging all its state info off  *
 * the GifFileType pointer.  Call DGifOpenFileName() or DGifOpenFileHandle()   *
 * first to initialize I/O.  Its inverse is EGifSpew().			      *
-******************************************************************************/
+* 
+ ******************************************************************************/
 int DGifSlurp(GifFileType *GifFile)
 {
     int i, j, Error, ImageSize;
     GifRecordType RecordType;
     SavedImage *sp;
-    ExtensionBlock *ep;
     GifByteType *ExtData;
+    SavedImage temp_save;
 
-    /* Some versions of malloc dislike 0-length requests */
-    GifFile->SavedImages = (SavedImage *)malloc(sizeof(SavedImage));
-
+    temp_save.ExtensionBlocks=NULL;
+    temp_save.ExtensionBlockCount=0;
+  
     do {
 	if (DGifGetRecordType(GifFile, &RecordType) == GIF_ERROR)
 	    return(GIF_ERROR);
@@ -899,32 +942,36 @@ int DGifSlurp(GifFileType *GifFile)
 		    == GIF_ERROR)
 		    return(GIF_ERROR);
 
+        if (temp_save.ExtensionBlocks) {
+            sp->ExtensionBlocks = temp_save.ExtensionBlocks;
+            sp->ExtensionBlockCount = temp_save.ExtensionBlockCount;
+
+            temp_save.ExtensionBlocks = NULL;
+            temp_save.ExtensionBlockCount=0;
+
+        /* FIXME: The following is wrong.  It is left in only for backwards
+         * compatibility.  Someday it should go away.  Use the
+         * sp->ExtensionBlocks->Function variable instead.
+         */
+            sp->Function = sp->ExtensionBlocks[0].Function;
+
+        }
+
 		break;
 
 	    case EXTENSION_RECORD_TYPE:
-
-		if (DGifGetExtension(GifFile,&sp->Function,&ExtData)==GIF_ERROR)
+		if (DGifGetExtension(GifFile,&temp_save.Function,&ExtData)==GIF_ERROR)
 		    return(GIF_ERROR);
-		else
-		{
-		    ep = &sp->ExtensionBlocks[sp->ExtensionBlockCount++];
-
-		    ep->ByteCount = ExtData[0];
-		    ep->Bytes = (GifByteType *)malloc(ep->ByteCount * sizeof(GifByteType));
-		    memcpy(ep->Bytes, ExtData, ep->ByteCount * sizeof(char));
-		}
-
 		while (ExtData != NULL) {
+            
+            /* Create an extension block with our data */
+            if (AddExtensionBlock(&temp_save, ExtData[0], &ExtData[1])
+               == GIF_ERROR)
+                return (GIF_ERROR); 
+            
 		    if (DGifGetExtensionNext(GifFile, &ExtData) == GIF_ERROR)
-			return(GIF_ERROR);
-		    else
-		    {
-			ep = &sp->ExtensionBlocks[sp->ExtensionBlockCount++];
-
-			ep->ByteCount = ExtData[0];
-			ep->Bytes = (GifByteType *)malloc(ep->ByteCount * sizeof(GifByteType));
-			memcpy(ep->Bytes,ExtData,ep->ByteCount * sizeof(char));
-		    }
+			    return(GIF_ERROR);
+            temp_save.Function = 0;
 		}
 		break;
 
@@ -934,10 +981,16 @@ int DGifSlurp(GifFileType *GifFile)
 	    default:	/* Should be trapped by DGifGetRecordType */
 		break;
 	}
-    }
-    while
-	(RecordType != TERMINATE_RECORD_TYPE);
+    } while (RecordType != TERMINATE_RECORD_TYPE);
 
+    /* Just in case the Gif has an extension block without an associated
+     * image... (Should we save this into a savefile structure with no image
+     * instead?  Have to check if the present writing code can handle that as
+     * well....
+     */
+    if (temp_save.ExtensionBlocks)
+        FreeExtension(&temp_save);
+    
     return(GIF_OK);
 }
 
