@@ -24,68 +24,28 @@
 #endif /* __MSDOS__ */
 
 #include <fcntl.h>
+#include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
 #include "gif_lib.h"
-#include "gif_hash.h"
-
-#define PROGRAM_NAME	"GIFLIB"
+#include "gif_lib_private.h"
 
 #define GIF87_STAMP	"GIF87a"         /* First chars in file - GIF stamp. */
 #define GIF89_STAMP	"GIF89a"         /* First chars in file - GIF stamp. */
 
-#define ZL_MAX_CODE	4095		/* Biggest code possible in 12 bits. */
-
 #define FILE_STATE_WRITE	0x01/* 1 write, 0 read - DGIF_LIB compatible.*/
 #define FILE_STATE_SCREEN	0x02
 #define FILE_STATE_IMAGE	0x04
 
-#define FLUSH_OUTPUT		4096	/* Impossible code, to signal flush. */
-#define FIRST_CODE		4097    /* Impossible code, to signal first. */
-
 #define IS_WRITEABLE(Private)	(Private->FileState & FILE_STATE_WRITE)
 
 /* #define DEBUG_NO_PREFIX		          Dump only compressed data. */
 
-typedef struct GifFilePrivateType {
-    int FileState,
-	FileHandle,			     /* Where old this data goes to! */
-	BitsPerPixel,	   /* Bits per pixel (Codes uses at least this + 1). */
-	ClearCode,				       /* The CLEAR LZ code. */
-	EOFCode,					 /* The EOF LZ code. */
-	RunningCode,		    /* The next code algorithm can generate. */
-	RunningBits,/* The number of bits required to represent RunningCode. */
-	MaxCode1,  /* 1 bigger than max. possible code, in RunningBits bits. */
-	CrntCode,				  /* Current algorithm code. */
-	CrntShiftState;			/* Number of bits in CrntShiftDWord. */
-    unsigned long CrntShiftDWord;     /* For bytes decomposition into codes. */
-    unsigned long PixelCount;
-    FILE *File;						  /* File as stream. */
-    GifByteType Buf[256];	      /* Compressed output is buffered here. */
-    GifHashTableType *HashTable;
-} GifFilePrivateType;
-
-extern int _GifError;
-
 /* Masks given codes to BitsPerPixel, to make sure all codes are in range: */
 static GifPixelType CodeMask[] = {
     0x00, 0x01, 0x03, 0x07, 0x0f, 0x1f, 0x3f, 0x7f, 0xff
 };
 
-#ifdef SYSV
-static char *VersionStr =
-        "Gif library module,\t\tGershon Elber\n\
-	(C) Copyright 1989 Gershon Elber, Non commercial use only.\n";
-#else
-static char *VersionStr =
-	PROGRAM_NAME
-	"	IBMPC "
-	GIF_LIB_VERSION
-	"	Gershon Elber,	"
-	__DATE__ ",   " __TIME__ "\n"
-	"(C) Copyright 1989 Gershon Elber, Non commercial use only.\n";
-#endif /* SYSV */
-
 static char *GifVersionPrefix = GIF87_STAMP;
 
 static int EGifPutWord(int Word, FILE *File);
@@ -157,15 +117,6 @@ GifFileType *EGifOpenFileHandle(int FileHandle)
 
     memset(GifFile, '\0', sizeof(GifFileType));
 
-#ifndef DEBUG_NO_PREFIX
-    if (fwrite(GifVersionPrefix, 1, strlen(GifVersionPrefix), f) !=
-						strlen(GifVersionPrefix)) {
-	_GifError = E_GIF_ERR_WRITE_FAILED;
-	free((char *) GifFile);
-	return NULL;
-    }
-#endif /* DEBUG_NO_PREFIX */
-
     if ((Private = (GifFilePrivateType *) malloc(sizeof(GifFilePrivateType)))
 	== NULL) {
 	_GifError = E_GIF_ERR_NOT_ENOUGH_MEM;
@@ -177,10 +128,11 @@ GifFileType *EGifOpenFileHandle(int FileHandle)
     Private->File = f;
     Private->FileState = FILE_STATE_WRITE;
     if ((Private->HashTable = _InitHashTable()) == NULL) {
-	_GifError = E_GIF_ERR_NOT_ENOUGH_MEM;
-	return NULL;
+       _GifError = E_GIF_ERR_NOT_ENOUGH_MEM;
+       return NULL;
     }
 
+
     _GifError = 0;
 
     return GifFile;
@@ -219,6 +171,15 @@ int EGifPutScreenDesc(GifFileType *GifFile,
 	return GIF_ERROR;
     }
 
+/* First write the version prefix into the file. */
+#ifndef DEBUG_NO_PREFIX
+    if (fwrite(GifVersionPrefix, 1, strlen(GifVersionPrefix), Private->File) !=
+						strlen(GifVersionPrefix)) {
+	_GifError = E_GIF_ERR_WRITE_FAILED;
+	return GIF_ERROR;
+    }
+#endif /* DEBUG_NO_PREFIX */
+
     GifFile->SWidth = Width;
     GifFile->SHeight = Height;
     GifFile->SColorResolution = ColorRes;
@@ -527,8 +488,8 @@ int EGifCloseFile(GifFileType *GifFile)
     if (GifFile->SColorMap)
 	FreeMapObject(GifFile->SColorMap);
     if (Private) {
-	if (Private->HashTable) free((char *) Private->HashTable);
-	free((char *) Private);
+        if (Private->HashTable) free((char *) Private->HashTable);
+	    free((char *) Private);
     }
     free(GifFile);
 
@@ -591,8 +552,9 @@ static int EGifSetupCompress(GifFileType *GifFile)
     Private->CrntShiftState = 0;      /* No information in CrntShiftDWord. */
     Private->CrntShiftDWord = 0;
 
-    /* Clear hash table and send Clear to make sure the decoder do the same. */
+   /* Clear hash table and send Clear to make sure the decoder do the same. */
     _ClearHashTable(Private->HashTable);
+
     if (EGifCompressOutput(Private, Private->ClearCode) == GIF_ERROR) {
 	_GifError = E_GIF_ERR_DISK_IS_FULL;
 	return GIF_ERROR;
@@ -624,43 +586,44 @@ static int EGifCompressLine(GifFileType *GifFile, GifPixelType *Line,
 
     while (i < LineLen) {			    /* Decode LineLen items. */
 	Pixel = Line[i++];		      /* Get next pixel from stream. */
-	/* Form a new unique key to search hash table for the code combines  */
-	/* CrntCode as Prefix string with Pixel as postfix char.	     */
-	NewKey = (((unsigned long) CrntCode) << 8) + Pixel;
-	if ((NewCode = _ExistsHashTable(HashTable, NewKey)) >= 0) {
-	    /* This Key is already there, or the string is old one, so	     */
-	    /* simple take new code as our CrntCode:			     */
-	    CrntCode = NewCode;
-	}
-	else {
-	    /* Put it in hash table, output the prefix code, and make our    */
-	    /* CrntCode equal to Pixel.					     */
-	    if (EGifCompressOutput(Private, CrntCode)
-		== GIF_ERROR) {
-		_GifError = E_GIF_ERR_DISK_IS_FULL;
-		return GIF_ERROR;
-	    }
-	    CrntCode = Pixel;
+       /* Form a new unique key to search hash table for the code combines  */
+       /* CrntCode as Prefix string with Pixel as postfix char.             */
+       NewKey = (((unsigned long) CrntCode) << 8) + Pixel;
+       if ((NewCode = _ExistsHashTable(HashTable, NewKey)) >= 0) {
+           /* This Key is already there, or the string is old one, so       */
+           /* simple take new code as our CrntCode:                         */
+           CrntCode = NewCode;
+       }
+       else {
+           /* Put it in hash table, output the prefix code, and make our    */
+           /* CrntCode equal to Pixel.                                      */
+           if (EGifCompressOutput(Private, CrntCode)
+               == GIF_ERROR) {
+               _GifError = E_GIF_ERR_DISK_IS_FULL;
+               return GIF_ERROR;
+           }
+           CrntCode = Pixel;
+
+           /* If however the HashTable if full, we send a clear first and   */
+           /* Clear the hash table.                                         */
+           if (Private->RunningCode >= LZ_MAX_CODE) {
+               /* Time to do some clearance: */
+               if (EGifCompressOutput(Private, Private->ClearCode)
+                   == GIF_ERROR) {
+                   _GifError = E_GIF_ERR_DISK_IS_FULL;
+                   return GIF_ERROR;
+               }
+               Private->RunningCode = Private->EOFCode + 1;
+               Private->RunningBits = Private->BitsPerPixel + 1;
+               Private->MaxCode1 = 1 << Private->RunningBits;
+               _ClearHashTable(HashTable);
+           }
+           else {
+               /* Put this unique key with its relative Code in hash table: */
+               _InsertHashTable(HashTable, NewKey, Private->RunningCode++);
+           }
+       }
 
-	    /* If however the HashTable if full, we send a clear first and   */
-	    /* Clear the hash table.					     */
-	    if (Private->RunningCode >= ZL_MAX_CODE) {
-		/* Time to do some clearance: */
-		if (EGifCompressOutput(Private, Private->ClearCode)
-		    == GIF_ERROR) {
-		    _GifError = E_GIF_ERR_DISK_IS_FULL;
-		    return GIF_ERROR;
-		}
-		Private->RunningCode = Private->EOFCode + 1;
-		Private->RunningBits = Private->BitsPerPixel + 1;
-		Private->MaxCode1 = 1 << Private->RunningBits;
-		_ClearHashTable(HashTable);
-	    }
-	    else {
-		/* Put this unique key with its relative Code in hash table: */
-		_InsertHashTable(HashTable, NewKey, Private->RunningCode++);
-	    }
-	}
     }
 
     /* Preserve the current state of the compression algorithm: */
@@ -728,7 +691,7 @@ static int EGifCompressOutput(GifFilePrivateType *Private, int Code)
     /* If code cannt fit into RunningBits bits, must raise its size. Note */
     /* however that codes above 4095 are used for special signaling.      */
     if (Private->RunningCode >= Private->MaxCode1 && Code <= 4095) {
-	Private->MaxCode1 = 1 << ++Private->RunningBits;
+       Private->MaxCode1 = 1 << ++Private->RunningBits;
     }
 
     return retval;
@@ -784,13 +747,15 @@ int EGifSpew(GifFileType *GifFileOut)
 
     for (i = 0; i < GifFileOut->ImageCount; i++)
     {
-	int function = GifFileOut->SavedImages[i].Function;
+        for (j = 0; j < GifFileOut->SavedImages[i].ExtensionBlockCount; j++) {
+            int function=GifFileOut->SavedImages[i].ExtensionBlocks[j].Function;
 
-	if (function == COMMENT_EXT_FUNC_CODE
-		    || function == GRAPHICS_EXT_FUNC_CODE
-		    || function == PLAINTEXT_EXT_FUNC_CODE
-		    || function == APPLICATION_EXT_FUNC_CODE)
-	    gif89 = TRUE;
+	        if (function == COMMENT_EXT_FUNC_CODE
+                || function == GRAPHICS_EXT_FUNC_CODE
+                || function == PLAINTEXT_EXT_FUNC_CODE
+                || function == APPLICATION_EXT_FUNC_CODE)
+            gif89 = TRUE;
+        }
     }
 
     SavedStamp = GifVersionPrefix;
@@ -817,6 +782,17 @@ int EGifSpew(GifFileType *GifFileOut)
 	/* this allows us to delete images by nuking their rasters */
 	if (sp->RasterBits == NULL)
 	    continue;
+	
+    if (sp->ExtensionBlocks)
+	{
+        for ( j = 0; j < sp->ExtensionBlockCount; j++) {
+            ep = &sp->ExtensionBlocks[j];
+            if (EGifPutExtension(GifFileOut,
+               (ep->Function != 0) ? ep->Function : '\0',
+               ep->ByteCount, ep->Bytes) == GIF_ERROR)
+                return (GIF_ERROR);
+        }
+    }
 
 	if (EGifPutImageDesc(GifFileOut,
 			     sp->ImageDesc.Left,
@@ -835,17 +811,6 @@ int EGifSpew(GifFileType *GifFileOut)
 			    SavedWidth) == GIF_ERROR)
 		return(GIF_ERROR);
 	}
-
-	if (sp->ExtensionBlocks)
-	{
-	    for (ep = sp->ExtensionBlocks;
-		 ep < sp->ExtensionBlocks + sp->ExtensionBlockCount;
-		 ep++)
-	    if (EGifPutExtension(GifFileOut,
-			     (ep == sp->ExtensionBlocks) ? sp->Function : '\0',
-			     ep->ByteCount, ep->Bytes) == GIF_ERROR)
-		return(GIF_ERROR);
-	}
     }
 
     if (EGifCloseFile(GifFileOut) == GIF_ERROR)
